# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:

* 文本特征：包括Query长度、Term长度，Term在Query中的偏移量，term词性、长度信息、
term数目、位置信息、句法依存tag、是否数字、是否英文、是否停用词、是否专名实体、
是否重要行业词、embedding模长、删词差异度、以及短语生成树得到term权重等。

"""

import codecs
from copy import deepcopy

import numpy as np
from text2vec import Vector, Similarity
from text2vec.similarity import SimType

from wordrank import config
from wordrank.utils.logger import logger
from wordrank.utils.text_utils import is_number_string, is_alphabet_string, is_chinese_string
from wordrank.utils.tokenizer import word_segment


class AttrDict(dict):
    """Dict that can get attribute by dot"""

    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


class TextFeature(object):
    def __init__(self,
                 stopwords_path=config.stopwords_path,
                 person_name_path=config.person_name_path,
                 place_name_path=config.place_name_path,
                 common_char_path=config.common_char_path,
                 segment_sep=config.segment_sep
                 ):
        self.stopwords = self.load_set_file(stopwords_path)
        self.person_names = self.load_set_file(person_name_path)
        self.place_names = self.load_set_file(place_name_path)
        self.common_chars = self.load_set_file(common_char_path)
        self.segment_sep = segment_sep
        self.vec = Vector()
        self.sim = Similarity(similarity_type=SimType.WMD)

    @staticmethod
    def load_set_file(path):
        words = set()
        with codecs.open(path, 'r', encoding='utf-8') as f:
            for w in f:
                w = w.strip()
                if w.startswith('#'):
                    continue
                if w:
                    words.add(w.split()[0])
        return words

    def is_stopword(self, word):
        return word in self.stopwords

    def is_name(self, word):
        names = self.person_names | self.place_names
        return word in names

    def is_entity(self, pos, entity_pos=('ns', 'n', 'vn', 'v')):
        return pos in entity_pos

    def is_common_char(self, c):
        return c in self.common_chars

    def is_common_char_string(self, word):
        return all(self.is_common_char(c) for c in word)

    def get_feature(self, query, is_word_segmented=False):
        """
        Get text feature
        :param query:
        :param is_word_segmented:
        :return: list, list: term features, sentence features
        """
        term_features = []
        if is_word_segmented:
            word_seq = query.split(self.segment_sep)
        else:
            word_seq = word_segment(query, cut_type='word', pos=False)
        logger.debug('%s' % word_seq)

        # sentence
        sentence_features = AttrDict(
            query_length=len(query),
            term_size=len(word_seq),
        )

        # term
        idx = 0
        offset = 0
        for word in word_seq:
            emb = self.vec.encode(word)
            word_list = deepcopy(word_seq)
            if word in word_list:
                word_list.remove(word)
            del_word_query = ''.join(word_list)
            del_term_sim_score = self.sim.get_score(query, del_word_query)
            term_features.append(AttrDict(
                term=word,
                term_length=len(word),
                idx=idx,
                offset=offset,
                is_number=is_number_string(word),
                is_chinese=is_chinese_string(word),
                is_alphabet=is_alphabet_string(word),
                is_stopword=self.is_stopword(word),
                is_name=self.is_name(word),
                # is_entity=self.is_entity(pos),
                is_common_char=self.is_common_char_string(word),
                embedding_sum=np.sum(emb),
                del_term_score=del_term_sim_score,
            ))
            idx += len(word)
            offset += 1

        return term_features, sentence_features



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""

import codecs

import jieba.posseg

from wordrank import config

class AttrDict(dict):
    """Dict that can get attribute by dot"""

    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self



class WordSegmentation(object):
    """ 分词 """

    def __init__(self, stop_words_file=config.stopwords_path, allow_speech_tags=config.allow_speech_tags):
        """
        Keyword arguments:
        stop_words_file    -- 保存停止词的文件路径，utf8编码，每行一个停止词。若不是str类型，则使用默认的停止词
        allow_speech_tags  -- 词性列表，用于过滤
        """

        allow_speech_tags = [item for item in allow_speech_tags]

        self.default_speech_tag_filter = allow_speech_tags
        self.stop_words = set()
        self.stop_words_file = stop_words_file
        if type(stop_words_file) is str:
            self.stop_words_file = stop_words_file
        for word in codecs.open(self.stop_words_file, 'r', 'utf-8', 'ignore'):
            self.stop_words.add(word.strip())

    def segment(self, text, lower=True, use_stop_words_filter=True, use_speech_tags_filter=False):
        """对一段文本进行分词，返回list类型的分词结果
        Keyword arguments:
        lower                  -- 是否将单词小写（针对英文）
        use_stop_words_filter         -- 若为True，则利用停止词集合来过滤（去掉停止词）
        use_speech_tags_filter -- 是否基于词性进行过滤。若为True，则使用self.default_speech_tag_filter过滤。否则，不过滤。
        """
        jieba_result = jieba.posseg.cut(text)

        if use_speech_tags_filter:
            jieba_result = [w for w in jieba_result if w.flag in self.default_speech_tag_filter]
        else:
            jieba_result = [w for w in jieba_result]

        # 去除特殊符号
        word_list = [w.word.strip() for w in jieba_result if w.flag != 'x']
        word_list = [word for word in word_list if len(word) > 0]

        if lower:
            word_list = [word.lower() for word in word_list]

        if use_stop_words_filter:
            word_list = [word.strip() for word in word_list if word.strip() not in self.stop_words]

        return word_list

    def segment_sentences(self, sentences, lower=True, use_stop_words_filter=True, use_speech_tags_filter=False):
        """将列表sequences中的每个元素/句子转换为由单词构成的列表。

        sequences -- 列表，每个元素是一个句子（字符串类型）
        """
        res = []
        for sentence in sentences:
            res.append(self.segment(text=sentence,
                                    lower=lower,
                                    use_stop_words_filter=use_stop_words_filter,
                                    use_speech_tags_filter=use_speech_tags_filter))
        return res


class SentenceSegmentation(object):
    """ 分句 """

    def __init__(self, delimiters=config.sentence_delimiters):
        """
        Keyword arguments:
        delimiters -- 可迭代对象，用来拆分句子
        """
        self.delimiters = set([item for item in delimiters])

    def segment(self, text):
        res = [text]

        for sep in self.delimiters:
            text, res = res, []
            for seq in text:
                res += seq.split(sep)
        res = [s.strip() for s in res if len(s.strip()) > 0]
        return res


class Segmentation(object):

    def __init__(self,
                 stopwords_file=config.stopwords_path,
                 allow_speech_tags=config.allow_speech_tags,
                 delimiters=config.sentence_delimiters):
        """
        Keyword arguments:
        stop_words_file -- 停止词文件
        delimiters      -- 用来拆分句子的符号集合
        """
        self.ws = WordSegmentation(stop_words_file=stopwords_file, allow_speech_tags=allow_speech_tags)
        self.ss = SentenceSegmentation(delimiters=delimiters)

    def segment(self, text, lower=False):
        sentences = self.ss.segment(text)
        words = self.ws.segment_sentences(sentences=sentences,
                                          lower=lower,
                                          use_stop_words_filter=False,
                                          use_speech_tags_filter=False)
        words_no_stop_words = self.ws.segment_sentences(sentences=sentences,
                                                        lower=lower,
                                                        use_stop_words_filter=True,
                                                        use_speech_tags_filter=False)

        words_all_filters = self.ws.segment_sentences(sentences=sentences,
                                                      lower=lower,
                                                      use_stop_words_filter=True,
                                                      use_speech_tags_filter=True)

        return AttrDict(
            sentences=sentences,
            words=words,
            words_no_stop_words=words_no_stop_words,
            words_all_filters=words_all_filters
        )



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:

前后词互信息PMI: 计算两个词语之间的关联程度，pmi越大越紧密
左右邻熵：左右熵值越大，说明该词的周边词越丰富，意味着词的自由程度越大，其成为一个独立的词的可能性也就越大。
"""

import os
import re
from collections import Counter

import numpy as np

from wordrank import config
from wordrank.utils.io_utils import save_json, load_json
from wordrank.utils.logger import logger


def ngram_freq(text, ngram=4):
    """
    Get Ngram freq dict
    :param text: input text, sentence
    :param ngram: N
    :return: word frequency dict
    """
    words = []
    for i in range(1, ngram + 1):
        words += [text[j:j + i] for j in range(len(text) - i + 1)]
    words_freq = dict(Counter(words))
    return words_freq


def generate_pmi_score(word_freq_dict, pmi_path):
    """
    Generate PMI score
    :param word_freq_dict: dict
    :return: dict(word, float), pmi score
    """
    result = dict()
    for word in word_freq_dict:
        if len(word) > 1:
            p_x_y = min([word_freq_dict.get(word[:i]) * word_freq_dict.get(word[i:]) for i in range(1, len(word))])
            score = p_x_y / word_freq_dict.get(word)
            result[word] = score
    save_json(result, pmi_path)
    return result


def entropy_score(char_list):
    """
    Get entropy score
    :param char_list: input char list
    :return: float, score
    """
    char_freq_dict = dict(Counter(char_list))
    char_size = len(char_list)
    entropy = -1 * sum([char_freq_dict.get(i) / char_size * np.log2(char_freq_dict.get(i) / char_size)
                        for i in char_freq_dict])
    return entropy


def generate_entropy_score(word_freq_dict, text, entropy_path):
    """
    Generate entropy score
    :param word_freq_dict: dict
    :param text: input text, document
    :param entropy_path: save entropy file path
    :return: dict
    """
    result = dict()
    for word in word_freq_dict:
        if len(word) == 1:
            continue
        # pass error pattern
        if '*' in word:
            continue
        try:
            left_right_char = re.findall('(.)%s(.)' % word, text)
            left_char = [i[0] for i in left_right_char]
            left_entropy = entropy_score(left_char)

            right_char = [i[1] for i in left_right_char]
            right_entropy = entropy_score(right_char)
            if left_entropy > 0 or right_entropy > 0:
                result[word] = [left_entropy, right_entropy]
        except Exception as e:
            logger.error('error word %s, %s' % (word, e))
    save_json(result, entropy_path)
    return result


class PMI(object):
    def __init__(self,
                 text='',
                 ngram=4,
                 pmi_path=config.pmi_path,
                 entropy_path=config.entropy_path
                 ):
        if os.path.exists(pmi_path) and os.path.exists(entropy_path):
            self.pmi_score_dict = load_json(pmi_path)
            self.entropy_score_dict = load_json(entropy_path)
        elif len(text.strip()) > 0:
            words_freq = ngram_freq(text, ngram)
            self.pmi_score_dict = generate_pmi_score(words_freq, pmi_path)
            self.entropy_score_dict = generate_entropy_score(words_freq, text, entropy_path)
        else:
            raise ValueError("text and saved file path must exist one")

    def pmi_score(self, word):
        """
        Get PMI score
        :param word:
        :return:
        """
        return self.pmi_score_dict.get(word, 0.0)

    def entropy_score(self, word):
        """
        Get entropy score
        :param word:
        :return:
        """
        return self.entropy_score_dict.get(word, [0.0, 0.0])



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: Word Rank module, main
"""

from wordrank import config
from wordrank.feature import Feature
from wordrank.utils import tokenizer
from wordrank.utils.io_utils import load_pkl
from wordrank.utils.logger import logger
from wordrank.utils.text_utils import convert_to_unicode


class WordRank(Feature):
    def __init__(self,
                 model_path=config.model_path,
                 stopwords_path=config.stopwords_path,
                 person_name_path=config.person_name_path,
                 place_name_path=config.place_name_path,
                 common_char_path=config.common_char_path,
                 segment_sep = config.segment_sep,
                 domain_sample_path=config.domain_sample_path,
                 ngram=config.ngram,
                 pmi_path=config.pmi_path,
                 entropy_path=config.entropy_path,
                 sentence_delimiters=config.sentence_delimiters,
                 ):
        super(WordRank, self).__init__(
            stopwords_path=stopwords_path,
            person_name_path=person_name_path,
            place_name_path=place_name_path,
            common_char_path=common_char_path,
            segment_sep = segment_sep,
            domain_sample_path=domain_sample_path,
            ngram=ngram,
            pmi_path=pmi_path,
            entropy_path=entropy_path,
            sentence_delimiters=sentence_delimiters
        )
        self.model_path = model_path
        self.inited = False

    def _init(self):
        self.model = load_pkl(self.model_path)
        self.inited = True

    def check_inited(self):
        if not self.inited:
            self._init()

    def rank_query(self, query):
        self.check_inited()
        # get feature
        data_feature, terms = self.get_feature(query, is_word_segmented=False)
        # predict model
        label_pred = self.model.predict(data_feature)
        logger.debug("predict label: %s" % label_pred)
        return zip(terms, label_pred)

    def rank(self, text):
        """
        Word Rank
        :param text:
        :return:
        """
        result = []
        self.check_inited()
        # unicode
        text = convert_to_unicode(text)
        # split to short sentence
        sentences = tokenizer.sentence_segment(text, delimiters=self.sentence_delimiters, include_symbols=True)
        for sentence, idx in sentences:
            for w, p in self.rank_query(sentence):
                result.append((w, p))
        return result



# -*- coding: utf-8 -*-
# Author: XuMing(xuming624@qq.com)
# Brief: 
from __future__ import print_function

import sys

from setuptools import setup, find_packages

__version__ = None
exec(open('wordrank/version.py').read())

if sys.version_info < (3,):
    sys.exit('Sorry, Python3 is required for wordrank.')

with open('README.md', 'r', encoding='utf-8') as f:
    readme = f.read()

with open('LICENSE', 'r', encoding='utf-8') as f:
    license = f.read()

with open('requirements.txt', 'r', encoding='utf-8') as f:
    reqs = f.read()

setup(
    name='wordrank',
    version=__version__,
    description='word rank, context-aware bag-of-words term weights for query and document.',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='XuMing',
    author_email='xuming624@qq.com',
    url='https://github.com/shibing624/wordrank',
    license="Apache 2.0",
    zip_safe=False,
    python_requires='>=3.5',
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Natural Language :: Chinese (Simplified)',
        'Natural Language :: Chinese (Traditional)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Text Processing :: Linguistic',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    platforms=["Windows", "Linux", "Solaris", "Mac OS-X", "Unix"],
    keywords='wordrank,term weighting,textrank,word rank',
    install_requires=reqs.strip().split('\n'),
    packages=find_packages(exclude=['tests']),
    package_dir={'wordrank': 'wordrank'},
    package_data={'wordrank': ['*.*', '../LICENSE', '../README.*', '../*.txt', 'data/*',
                               '../extra_data/*.*']}
)



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""

from wordrank import config
from wordrank.features.language_feature import LanguageFeature
from wordrank.features.statistics_feature import StatisticsFeature
from wordrank.features.text_feature import TextFeature
from wordrank.utils.logger import logger


class Feature(object):
    def __init__(self,
                 stopwords_path=config.stopwords_path,
                 person_name_path=config.person_name_path,
                 place_name_path=config.place_name_path,
                 common_char_path=config.common_char_path,
                 segment_sep=config.segment_sep,
                 domain_sample_path=config.domain_sample_path,
                 ngram=config.ngram,
                 pmi_path=config.pmi_path,
                 entropy_path=config.entropy_path,
                 sentence_delimiters=config.sentence_delimiters,
                 ):
        self.stopwords_path = stopwords_path
        self.person_name_path = person_name_path
        self.place_name_path = place_name_path
        self.common_char_path = common_char_path
        self.segment_sep = segment_sep

        self.domain_sample_path = domain_sample_path
        self.ngram = ngram
        self.pmi_path = pmi_path
        self.entropy_path = entropy_path
        self.sentence_delimiters = sentence_delimiters
        self.feature_inited = False

    def _init_feature(self):
        """
        load data
        :return:
        """
        self.text_feature = TextFeature(
            stopwords_path=self.stopwords_path,
            person_name_path=self.person_name_path,
            place_name_path=self.place_name_path,
            common_char_path=self.common_char_path,
            segment_sep=self.segment_sep
        )
        self.statistics_feature = StatisticsFeature(
            domain_sample_path=self.domain_sample_path,
            ngram=self.ngram,
            pmi_path=self.pmi_path,
            entropy_path=self.entropy_path)
        self.language_feature = LanguageFeature()
        self.feature_inited = True

    def check_feature_inited(self):
        """
        check if data loaded
        :return:
        """
        if not self.feature_inited:
            self._init_feature()

    def get_feature(self, query, is_word_segmented=False):
        """
        Get feature from query
        :param query: input query
        :param is_word_segmented: bool, is word segmented or not
        :return: features, terms
        """
        features = []
        terms = []

        self.check_feature_inited()
        text_terms, text_sent = self.text_feature.get_feature(query, is_word_segmented=is_word_segmented)
        stat_terms, stat_sent = self.statistics_feature.get_feature(query, is_word_segmented=is_word_segmented)
        lang_terms, lang_sent = self.language_feature.get_feature(query, is_word_segmented=is_word_segmented)
        # sentence feature
        text_sent.update(stat_sent)
        text_sent.update(lang_sent)
        logger.debug('sentence features: %s' % text_sent)
        sent_feature = [text_sent.query_length, text_sent.term_size, text_sent.ppl]
        # term feature
        for text, stat, lang in zip(text_terms, stat_terms, lang_terms):
            text.update(stat)
            text.update(lang)
            # logger.debug('term features: %s' % text)
            term_feature = [text.term_length, text.idx, text.offset, float(text.is_number),
                            float(text.is_chinese), float(text.is_alphabet), float(text.is_stopword),
                            float(text.is_name), float(text.is_common_char), text.embedding_sum, text.del_term_score,
                            text.idf, text.text_rank_score, text.tfidf_score, text.pmi_score, text.left_entropy_score,
                            text.right_entropy_score, text.del_term_ppl, text.term_ngram_score, text.left_term_score,
                            text.right_term_score]
            feature = sent_feature + term_feature
            features.append(feature)
            terms.append(text.term)
        logger.debug("[query]feature size: %s, term size: %s" % (len(features), len(terms)))

        return features, terms



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:

* 语言模型特征：整个query的语言模型概率/去掉该Term后的Query的语言模型概率。
"""

from copy import deepcopy

import pycorrector

from wordrank.features.text_feature import AttrDict
from wordrank.utils.tokenizer import word_segment


class LanguageFeature(object):
    def __init__(self, segment_sep=' '):
        self.segment_sep = segment_sep

    def get_ngram_score(self, words):
        """
        取n元文法得分
        :param words: list, 以词或字切分
        :param mode:
        :return:
        """
        return pycorrector.ngram_score(' '.join(words))

    def get_ppl_score(self, words):
        """
        取语言模型困惑度得分，越小句子越通顺
        :param words: list, 以词或字切分
        :param mode:
        :return:
        """
        return pycorrector.ppl_score(' '.join(words))

    def get_feature(self, query, is_word_segmented=False):
        """
        Get language feature
        :param query:
        :param is_word_segmented:
        :return: list, list: term features, sentence features
        """
        term_features = []
        if is_word_segmented:
            word_seq = query.split(self.segment_sep)
        else:
            word_seq = word_segment(query, cut_type='word', pos=False)

        # sentence
        sentence_features = AttrDict(
            ppl=self.get_ppl_score(word_seq),
        )

        # term
        count = 0
        for word in word_seq:
            word_list = deepcopy(word_seq)
            if word in word_list:
                word_list.remove(word)
            left_word = word_seq[count - 1] if count > 0 else ''
            right_word = word_seq[count + 1] if count < len(word_seq) - 1 else ''

            term_features.append(AttrDict(
                del_term_ppl=self.get_ppl_score(word_list),
                term_ngram_score=self.get_ngram_score(word),
                left_term_score=self.get_ngram_score(left_word + word),
                right_term_score=self.get_ngram_score(word + right_word)
            ))
            count += 1
        return term_features, sentence_features
